unit connection.login;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils,
  // MySQL units
  mysql56conn, sqldb, DB,
  // cryptography units
  openssl,
  // project units
  Connection, Connection.Types;

type

  { TLoginConnection }

  TLoginConnection = class(TConnection)
  private
    function GenerateSalt: TSalt;
  public
    // login functionalities
    function Login(const EMail: TEmail; const Password: TPassword): TLoginToken;
    function GetPasswordHint(const EMail: TEmail): string;
    function RegisterAccount(const EMail: TEmail;
      const Passsword: TPassword; const Hint: THint): TLoginToken;
    procedure LogOut(const Token: TLoginToken);
    procedure ChangePassword(const OldPassword, NewPassword: TPassword;
      const Hint: THint; const Token: TLoginToken);

  end;

implementation

// default user status for new users
const
  DefaultStatus = 100;

{ TLoginConnection }

// generates a 32 Byte random number as Salt
function TLoginConnection.GenerateSalt: TSalt;
var
  b: array[0..31] of byte;
begin
  // generate cryptographically secure byte string
  RAND_pseudo_bytes(@b[0], 32);
  // return as HEX string
  SetLength(Result, 64);
  BinToHex(@b[0], @Result[1], 32);
end;

// Checks for correct login information and generates a token if its valid
function TLoginConnection.Login(const EMail: TEmail;
  const Password: TPassword): TLoginToken;
var
  Hash, CompareHash: TPasswordHash;
  Salt: TSalt;
  UID: TUserID;
begin
  // Throw exception if not connected
  CheckConn;
  try
    // Setup query to get account information
    FMySQLQuery.SQL.Text :=
      'SELECT ID, Password, Salt FROM Accounts WHERE EMail = :email';
    // fill in params
    FMySQLQuery.ParamByName('email').AsString := EMail;
    // execute query and recieve results
    FMySQLQuery.Open;

    // check if exectly one user was found
    if FMySQLQuery.RecordCount <> 1 then
      raise ELoginException.Create('invalid login data');

    // get information from query
    Salt := FMySQLQuery.FieldByName('Salt').AsString;
    CompareHash := FMySQLQuery.FieldByName('Password').AsString;
    UID := FMySQLQuery.FieldByName('ID').AsLargeInt;

  finally
    // close query and commit transaction
    FMySQLQuery.Close;
    FMySQLTransaction.Commit;
  end;

  // generate password hash from salt and password
  Hash := SHA256Hash(Salt + Password);

  //check if the hashes match
  if Hash = CompareHash then
    Result := GenerateToken(UID) // create new session token
  else // throw exception otherwise
    raise ELoginException.Create('invalid login data');
end;

// gets password hint for a user with a given email
function TLoginConnection.GetPasswordHint(const EMail: TEmail): string;
begin
  // Throw exception if not connected
  CheckConn;
  try
    // setup query to get the hint from the email adress
    FMySQLQuery.SQL.Text := 'SELECT Hint FROM Accounts WHERE EMail = :email';
    // fill in params
    FMySQLQuery.ParamByName('email').AsString := EMail;
    // execute query and recieve results
    FMySQLQuery.Open;

    // check if exactly one account was found
    if FMySQLQuery.RecordCount <> 1 then
      raise ELoginException.Create('invalid email adress');

    // return Hint
    Result := FMySQLQuery.FieldByName('Hint').AsString;

  finally
    // Close query and commit transaction
    FMySQLQuery.Close;
    FMySQLTransaction.Commit;
  end;
end;

// creates new account and creates a session token for it
function TLoginConnection.RegisterAccount(const EMail: TEmail;
  const Passsword: TPassword; const Hint: THint): TLoginToken;
var
  Hash: TPasswordHash;
  Salt: TSalt;
  UID: TUserID;
begin
  // check for valid email adress
  if (Pos('@', EMail) < 2) or (pos('@', EMail) = Length(EMail)) then
    raise ELoginException.Create('invalid email adress');

  // check if password is not empty
  if Length(Passsword) = 0 then
    raise ELoginException.Create('invalid password');

  // Throw exception if not connected
  CheckConn;
  try
    // setup query to add useraccount to database
    FMySQLQuery.SQL.Text :=
      'INSERT INTO Accounts (EMail, Password, Salt, Hint, Status) Values ' +
      '(:email, :pass, :salt, :hint, :status)';

    // generate new salt
    Salt := GenerateSalt;
    // generate password hash from salt and password
    Hash := SHA256Hash(Salt + Passsword);

    // fill in params
    FMySQLQuery.ParamByName('email').AsString := EMail;
    FMySQLQuery.ParamByName('pass').AsString := Hash;
    FMySQLQuery.ParamByName('salt').AsString := Salt;
    FMySQLQuery.ParamByName('hint').AsString := Hint;
    FMySQLQuery.ParamByName('status').AsInteger := DefaultStatus;

    // execute sql
    FMySQLQuery.ExecSQL;

    // get last inserted ID
    UID := FMySQLConnection.GetInsertID;

    // setup query to add additional userdata
    FMySQLQuery.SQL.Text :=
      'INSERT INTO Users (UID, Username, FirstName, LastName) Values (:uid, '', '', '')';
    // fill in params
    FMySQLQuery.ParamByName('uid').AsLargeInt := UID;
    // execute query
    FMySQLQuery.ExecSQL;

    // Generate new session token and return it
    Result := GenerateToken(UID);

  finally
    // commit transaction
    FMySQLTransaction.Commit;
  end;
end;

// deletes session token from database
procedure TLoginConnection.LogOut(const Token: TLoginToken);
var
  UID: TUserID;
begin
  // get uid of the session referenced by token
  UID := CheckToken(Token);
  // if no session was found throw an exception
  if UID = 0 then
    raise EInvalidTokenException.Create('invalid token');

  try
    // setup query to delete session token
    FMySQLQuery.SQL.Text := 'DELETE FROM Tokens WHERE UID=:uid';
    // fill in params
    FMySQLQuery.ParamByName('uid').AsLargeInt := UID;
    // execute query
    FMySQLQuery.ExecSQL;
  finally
    // commit transaction
    FMySQLTransaction.Commit;
  end;
end;

// changes the users password, requires old password to match
procedure TLoginConnection.ChangePassword(const OldPassword, NewPassword: TPassword;
  const Hint: THint; const Token: TLoginToken);
var
  Salt, NewSalt: TSalt;
  Hash, NewHash, CompareHash: TPasswordHash;
  UID: TUserID;
begin
  // get uid of the session referenced by token
  UID := CheckToken(Token);
  // if no session was found throw an exception
  if UID = 0 then
    raise EInvalidTokenException.Create('invalid token');

  try
    // setup query to gather account information
    FMySQLQuery.SQL.Text := 'SELECT ID, Password, Salt FROM Accounts WHERE ID=:uid';
    // fill in params
    FMySQLQuery.ParamByName('uid').AsLargeInt := UID;
    // execute query and get results
    FMySQLQuery.Open;
    try
      // check if exactly one user was found (if not something gone terribly wrong)
      if FMySQLQuery.RecordCount <> 1 then
        raise EThisShouldNeverHappenException.Create('Recordcount=' +
          IntToStr(FMySQLQuery.RecordCount) + ', this should never happen');
      // get data from query
      Salt := FMySQLQuery.FieldByName('Salt').AsString;
      CompareHash := FMySQLQuery.FieldByName('Password').AsString;
    finally
      // close query
      FMySQLQuery.Close;
    end;

    // generate hash of the current password
    Hash := SHA256Hash(Salt + OldPassword);

    // compare if the passwords match
    if Hash <> CompareHash then
      raise ELoginException.Create('invalid password');

    // generate new salt for new password
    NewSalt := GenerateSalt;
    // generate hash of the new password
    NewHash := SHA256Hash(NewSalt + NewPassword);

    // setup query to update users password and salt
    FMySQLQuery.SQL.Text := 'UPDATE Accounts SET Password=:pass, Salt=:salt,' +
      ' Hint=:hint WHERE ID=:uid';
    // fill in params
    FMySQLQuery.ParamByName('pass').AsString := NewHash;
    FMySQLQuery.ParamByName('salt').AsString := NewSalt;
    FMySQLQuery.ParamByName('hint').AsString := Hint;
    FMySQLQuery.ParamByName('uid').AsLargeInt := UID;

    // execute query
    FMySQLQuery.ExecSQL;

  finally
    //commit transaction
    FMySQLTransaction.Commit;
  end;
end;

end.
