unit Connection;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils,
  // SQL database units
  mysql56conn, sqldb, DB,
  // cryptography units
  openssl, DCPsha256,
  // project units
  Connection.Types;

type
  { TConnection }

  TConnection = class
  private
    // IP Adress (used for hashs)
    FIP: TIPAdress;
    function TokenExsists(const RND: TLoginToken): boolean;
    function GenRND: TLoginToken;
  protected
    // MySQL objects
    FMySQLConnection: TMySQLConnection;
    FMySQLTransaction: TSQLTransaction;
    FMySQLQuery: TSQLQuery;
    // Basic Token functions
    procedure CheckConn; inline;
    function SHA256Hash(const Data: string): TSHA256Hash;
    function CheckToken(const Token: TLoginToken): TUserID; virtual;
    function GenerateToken(const ID: TUserID): TLoginToken; virtual;
  public
    constructor Create(AIP: TIPAdress);
    procedure Connect(DatabaseInfo: TDBInfo); virtual;
    destructor Destroy; override;

    property IP: TIPAdress read FIP write FIP;
  end;

  { TSessionConnection }

  TSessionConnection = class(TConnection)
  private
    FToken: TLoginToken;
    FUID: TUserID;
  public
    constructor Create(AIP: TIPAdress; AToken: TLoginToken);
    procedure Connect(DatabaseInfo: TDBInfo); override;

    property SessionToken: TLoginToken read FToken write FToken;
    property SessionUser: TUserID read FUID;
  end;

implementation

{ TSessionConnection }

constructor TSessionConnection.Create(AIP: TIPAdress; AToken: TLoginToken);
begin
  inherited Create(AIP);
  FToken:=AToken;
end;

procedure TSessionConnection.Connect(DatabaseInfo: TDBInfo);
begin
  inherited Connect(DatabaseInfo);
  // resolve session cookie
  FUID:=CheckToken(FToken);
  // check if valid session, else throw execption
  if FUID = 0 then
    raise EInvalidTokenException.Create('invalid token');
end;

{ TConnection }

// Throws an Exception if not connected to the database
procedure TConnection.CheckConn;
begin
  if not FMySQLConnection.Connected then
    raise EDBConnectionException.Create(
      'not connected to any database, call Connect first');
end;

// Generates a 32 Byte random number as login tokens
function TConnection.GenRND: TLoginToken;
var
  b: array[0..31] of byte;
begin
  // Cryptographically secure random byte string (openssl)
  RAND_pseudo_bytes(@b[0], 32);
  // write into Result as HEX
  SetLength(Result, 64);
  BinToHex(@b[0], @Result[1], 32);
end;

// Hashing function
function TConnection.SHA256Hash(const Data: string): TSHA256Hash;
var
  b: array[0..31] of byte;
begin
  with TDCP_sha256.Create(nil) do
    try
      Init;
      UpdateStr(Data);
      Final(b[0]);
      SetLength(Result, 64);
      BinToHex(@b[0], @Result[1], 32);
    finally
      Free;
    end;
end;

// Checks weather a token using this random number already exsists
function TConnection.TokenExsists(const RND: TLoginToken): boolean;
begin
  // Throw exception if not connected
  CheckConn;
  try
    // Setup query to get all sessions using this token
    FMySQLQuery.SQL.Text := 'SELECT RND FROM Tokens WHERE RND = :rnd';
    // fill in params
    FMySQLQuery.ParamByName('rnd').AsString := RND;
    // execute and recieve results
    FMySQLQuery.Open;
    // check if some was found
    Result := FMySQLQuery.RecordCount > 0;
  finally
    // close query (for next usage)
    FMySQLQuery.Close;
    // commit transaction
    FMySQLTransaction.Commit;
  end;
end;

// Checks for a token if it is valid, and if so return UID of the session
function TConnection.CheckToken(const Token: TLoginToken): TUserID;
var
  CompareHash, Hash: TSHA256Hash;
begin
  // Throw exception if not connected
  CheckConn;
  // Setup query to get all sessions using this token
  FMySQLQuery.SQL.Text :=
    'SELECT UID, Token, Expires FROM Tokens WHERE RND = :rnd ' +
    'AND NOW() < Expires';
  // fill in params
  FMySQLQuery.ParamByName('rnd').AsString := Token;
  // execute and recieve results
  FMySQLQuery.Open;
  try
    try
      // check if only one was found
      if FMySQLQuery.RecordCount <> 1 then
      begin
        Result := 0;
        Exit;
      end;

      // Read data from query
      CompareHash := FMySQLQuery.FieldByName('Token').AsString;
      Result := FMySQLQuery.FieldByName('UID').AsLargeInt;

      // generate token (Hash of IP+Randomnumber)
      Hash := SHA256Hash(IP + Token);

      // if not matching (wrong ip) discard
      if Hash <> CompareHash then
      begin
        Result := 0;
        Exit;
      end;

    finally
      // closing query for reuse
      FMySQLQuery.Close;
    end;

    // setup query to update the sessions time span
    FMySQLQuery.SQL.Text :=
      'UPDATE Tokens SET Expires=ADDTIME(NOW(), "0:30:00") WHERE UID = :uid';
    // fill in params
    FMySQLQuery.ParamByName('uid').AsLargeInt := Result;
    //execute
    FMySQLQuery.ExecSQL;
  finally
    // commit transaction
    FMySQLTransaction.Commit;
  end;
end;

// generates a token for a usersession
function TConnection.GenerateToken(const ID: TUserID): TLoginToken;
var
  RND, Hash: TSHA256Hash;
begin
  // Throw exception if not connected
  CheckConn;
  // generate non used 32 bit random number
  repeat
    RND := GenRND;
  until not TokenExsists(RND);

  // generate token (Hash of IP+Randomnumber)
  Hash := SHA256Hash(IP + RND);

  try
    // Setup query to add the new token, or update the timespan of exsisting one
    FMySQLQuery.SQL.Text :=
      'INSERT INTO Tokens (UID, Token, RND, Expires) VALUES ' +
      '(:uid,:tok,:rnd, ADDTIME(NOW(), "0:30:00")) ON DUPLICATE KEY ' +
      'UPDATE Token=:tok, RND=:rnd, Expires=ADDTIME(NOW(), "0:30:00")';
    // fill in params
    FMySQLQuery.ParamByName('rnd').AsString := RND;
    FMySQLQuery.ParamByName('uid').AsLargeInt := ID;
    FMySQLQuery.ParamByName('tok').AsString := Hash;

    // execute
    FMySQLQuery.ExecSQL;

  finally
    // commit transaction
    FMySQLTransaction.Commit;
  end;

  Result := RND;
end;

constructor TConnection.Create(AIP: TIPAdress);
begin
  // create MySQL objects
  FMySQLConnection := TMySQLConnection.Create(nil);
  FMySQLTransaction := TSQLTransaction.Create(nil);
  FMySQLQuery := TSQLQuery.Create(nil);
  // setup bindings between them
  FMySQLConnection.Transaction := FMySQLTransaction;
  FMySQLTransaction.DataBase := FMySQLConnection;
  FMySQLQuery.DataBase := FMySQLConnection;
  FMySQLQuery.Transaction := FMySQLTransaction;
  // init IP
  FIP := IP;
end;

// Setup instances
procedure TConnection.Connect(DatabaseInfo: TDBInfo);
begin
  // Setup connection
  with FMySQLConnection do
  begin
    HostName := DatabaseInfo.Hostname;
    UserName := DatabaseInfo.Username;
    Password := DatabaseInfo.Password;
    DatabaseName := DatabaseInfo.Database;
  end;

  // Open connection
  FMySQLConnection.Open;
  // throw error if connection is not available
  if not FMySQLConnection.Connected then
    raise EDBConnectionException.Create('Unable to connect to DB');
end;

// Cleanup instances
destructor TConnection.Destroy;
begin
  FMySQLConnection.Free;
  FMySQLTransaction.Free;
  FMySQLQuery.Free;
  inherited Destroy;
end;

end.
